"""Client for interacting with Airbyte instance"""

import logging
from typing import Tuple

import httpx

from prefect_airbyte import exceptions as err


class AirbyteClient:
    """
    Client class used to call API endpoints on an Airbyte server.

    This client assumes that you're using an OSS Airbyte server which does not require
    an API key to access its a API.

    For more info, see the [Airbyte docs](https://docs.airbyte.io/api-documentation).

    Attributes:
        logger: A logger instance used by the client to log messages related to
            API calls.
        airbyte_base_url str: Base API endpoint URL for Airbyte.
        timeout: The number of seconds to wait before an API call times out.
    """

    def __init__(
        self,
        logger: logging.Logger,
        airbyte_base_url: str = "http://localhost:8000/api/v1",
        timeout: int = 5,
    ):
        self.airbyte_base_url = airbyte_base_url
        self.logger = logger
        self.timeout = timeout

    async def _establish_session(self) -> httpx.AsyncClient:
        """
        Checks health of the Airbyte server and establishes a session.

        Returns:
            Session used to communicate with the Airbyte API.
        """
        client = httpx.AsyncClient(timeout=self.timeout)
        await self.check_health_status(client)
        return client

    async def check_health_status(self, client: httpx.AsyncClient) -> bool:
        """
        Checks the health status of an AirbyteInstance.

        Args:
            Session used to interact with the Airbyte API.

        Returns:
            True if the server is healthy. False otherwise.
        """
        get_connection_url = self.airbyte_base_url + "/health/"
        try:
            response = await client.get(get_connection_url)
            response.raise_for_status()

            self.logger.debug("Health check response: %s", response.json())
            key = "available" if "available" in response.json() else "db"
            health_status = response.json()[key]
            if not health_status:
                raise err.AirbyteServerNotHealthyException(
                    f"Airbyte Server health status: {health_status}"
                )
            return True
        except httpx.HTTPStatusError as e:
            raise err.AirbyteServerNotHealthyException() from e

    async def create_client(self) -> httpx.AsyncClient:
        """
        Convenience method for establishing a healthy session with the Airbyte server.

        Returns:
            Session for interacting with the Airbyte server.
        """
        client = await self._establish_session()
        return client

    async def export_configuration(
        self,
    ) -> bytes:
        """
        Triggers an export of Airbyte configuration.

        Returns:
            Gzipped Airbyte configuration data.
        """
        client = await self.create_client()

        get_connection_url = self.airbyte_base_url + "/deployment/export/"

        try:
            response = await client.post(get_connection_url)
            response.raise_for_status()

            self.logger.debug("Export configuration response: %s", response)

            export_config = response.content
            return export_config
        except httpx.HTTPStatusError as e:
            raise err.AirbyteExportConfigurationFailed() from e

    async def get_connection_status(self, connection_id: str) -> str:
        """
        Gets the status of a defined Airbyte connection.

        Args:
            connection_id: ID of an existing Airbyte connection.

        Returns:
            The status of the defined Airbyte connection.
        """
        client = await self.create_client()

        get_connection_url = self.airbyte_base_url + "/connections/get/"

        # TODO - Missing auth because Airbyte API currently doesn't yet support auth
        try:
            response = await client.post(
                get_connection_url, json={"connectionId": connection_id}
            )

            response.raise_for_status()

            connection_status = response.json()["status"]
            return connection_status
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                raise err.ConnectionNotFoundException() from e
            else:
                raise err.AirbyteServerNotHealthyException() from e

    async def trigger_manual_sync_connection(
        self, connection_id: str
    ) -> Tuple[str, str]:
        """
        Triggers a manual sync of the connection.

        Args:
            connection_id: ID of connection to sync.

        Returns:
            job_id: ID of the job that was triggered.
            created_at: Datetime string of when the job was created.

        """
        client = await self.create_client()

        get_connection_url = self.airbyte_base_url + "/connections/sync/"

        # TODO - no current authentication methods from Airbyte
        try:
            response = await client.post(
                get_connection_url, json={"connectionId": connection_id}
            )
            response.raise_for_status()
            job = response.json()["job"]
            job_id = job["id"]
            job_created_at = job["createdAt"]
            return job_id, job_created_at
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                raise err.ConnectionNotFoundException(
                    f"Connection {connection_id} not found, please double "
                    f"check the connection_id."
                ) from e

            raise err.AirbyteServerNotHealthyException() from e

    async def get_job_status(self, job_id: str) -> Tuple[str, str, str]:
        """
        Gets the status of an Airbyte connection sync job.

        Args:
            job_id: ID of the Airbyte job to check.

        Returns:
            job_status: The current status of the job.
            job_created_at: Datetime string of when the job was created.
            job_updated_at: Datetime string of the when the job was last updated.
        """
        client = await self.create_client()

        get_connection_url = self.airbyte_base_url + "/jobs/get/"
        try:
            response = await client.post(get_connection_url, json={"id": job_id})
            response.raise_for_status()

            job = response.json()["job"]
            job_status = job["status"]
            job_created_at = job["createdAt"]
            job_updated_at = job["updatedAt"]
            return job_status, job_created_at, job_updated_at
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                raise err.JobNotFoundException(f"Job {job_id} not found.") from e
            raise err.AirbyteServerNotHealthyException() from e



"""Exceptions to raise indicating issues throughout prefect_airbyte"""


class ConnectionNotFoundException(Exception):
    """
    Raises when a requested Airbyte connection cannot be found.
    """

    pass


class AirbyteServerNotHealthyException(Exception):
    """
    Raises when a specified Airbyte instance returns an unhealthy response.
    """

    pass


class JobNotFoundException(Exception):
    """
    Raises when a requested Airbyte job cannot be found.
    """

    pass


class AirbyteSyncJobFailed(Exception):
    """
    Raises when a specified Airbyte Sync Job fails.
    """

    pass


class AirbyteExportConfigurationFailed(Exception):
    """
    Raises when an Airbyte configuration export fails.
    """

    pass


class AirbyteConnectionInactiveException(Exception):
    """
    Raises when a specified Airbyte connection is inactive.
    """

    pass


class AirbyeConnectionDeprecatedException(Exception):
    """
    Raises when a specified Airbyte connection is deprecated.
    """

    pass



"""Tasks for connecting to Airbyte and triggering connection syncs"""
import uuid
from asyncio import sleep

from prefect import task
from prefect.logging.loggers import get_logger

from prefect_airbyte import exceptions as err
from prefect_airbyte.client import AirbyteClient

# Connection statuses
CONNECTION_STATUS_ACTIVE = "active"
CONNECTION_STATUS_INACTIVE = "inactive"
CONNECTION_STATUS_DEPRECATED = "deprecated"

# Job statuses
JOB_STATUS_SUCCEEDED = "succeeded"
JOB_STATUS_FAILED = "failed"
JOB_STATUS_PENDING = "pending"


@task
async def trigger_sync(
    airbyte_server_host: str = "localhost",
    airbyte_server_port: int = "8000",
    airbyte_api_version: str = "v1",
    connection_id: str = None,
    poll_interval_s: int = 15,
    status_updates: bool = False,
    timeout: int = 5,
) -> dict:
    """
    Task run method for triggering an Airbyte Connection.

    *It is assumed that the user will have previously configured
    a Source & Destination into a Connection.*
    e.g. MySql -> CSV

    An invocation of `trigger_sync` will attempt to start a sync job for
    the specified `connection_id` representing the Connection in
    Airbyte.

    `trigger_sync` will poll Airbyte Server for the Connection status and
    will only complete when the sync has completed or
    when it receives an error status code from an API call.

    Args:
        str airbyte_server_host : Hostname of Airbyte server where connection is
            configured.
        str airbyte_server_port: Port that the Airbyte server is listening on.

        str airbyte_api_version: Version of Airbyte API to use to trigger connection
            sync.
        str connection_id: the Airbyte connection ID
        int poll_interval_s: how often to poll the
            Airbyte API for sync status, if provided this will
            override the default polling time of 15 seconds.
        bool status_updates: whether to log status as the task polls jobs
        str timeout: The request `timeout` for the `httpx.AsyncClient`

    Returns:
        dict: connection_id (str) and succeeded_at (timestamp str)

    Examples:

        Flow that triggers an Airybte connection sync:

        ```python
        from prefect import flow
        from prefect_airbyte.connections import trigger_sync


        @flow
        def example_trigger_sync_flow():

            # Run other tasks and subflows here

            trigger_sync(
                connection_id="your-connection-id-to-sync"
            )

        example_trigger_sync_flow()
        ```

    """
    logger = get_logger()

    if not connection_id:
        raise ValueError(
            "Value for parameter `connection_id` *must* \
        be provided."
        )

    try:
        uuid.UUID(connection_id)
    except (TypeError, ValueError):
        raise ValueError(
            "Parameter `connection_id` *must* be a valid UUID \
            i.e. 32 hex characters, including hyphens."
        )

    # see https://airbyte-public-api-docs.s3.us-east-2.amazonaws.com
    # /rapidoc-api-docs.html#overview
    airbyte_base_url = (
        f"http://{airbyte_server_host}:"
        f"{airbyte_server_port}/api/{airbyte_api_version}"
    )

    airbyte = AirbyteClient(logger, airbyte_base_url, timeout=timeout)

    logger.info(
        f"Getting Airbyte Connection {connection_id}, poll interval "
        f"{poll_interval_s} seconds, airbyte_base_url {airbyte_base_url}"
    )

    connection_status = await airbyte.get_connection_status(connection_id)

    if connection_status == CONNECTION_STATUS_ACTIVE:
        # Trigger manual sync on the Connection ...
        job_id, job_created_at = await airbyte.trigger_manual_sync_connection(
            connection_id
        )

        job_status = JOB_STATUS_PENDING

        while job_status not in [JOB_STATUS_FAILED, JOB_STATUS_SUCCEEDED]:
            job_status, job_created_at, job_updated_at = await airbyte.get_job_status(
                job_id
            )

            # pending┃running┃incomplete┃failed┃succeeded┃cancelled
            if job_status == JOB_STATUS_SUCCEEDED:
                logger.info(f"Job {job_id} succeeded.")
            elif job_status == JOB_STATUS_FAILED:
                logger.error(f"Job {job_id} failed.")
                raise err.AirbyteSyncJobFailed(f"Job {job_id} failed.")
            else:
                if status_updates:
                    logger.info(job_status)
                # wait for next poll interval
                await sleep(poll_interval_s)

        return {
            "connection_id": connection_id,
            "status": connection_status,
            "job_status": job_status,
            "job_created_at": job_created_at,
            "job_updated_at": job_updated_at,
        }
    elif connection_status == CONNECTION_STATUS_INACTIVE:
        logger.error(
            f"Connection: {connection_id} is inactive"
            " - you'll need to enable it in your Airbyte instance"
        )
        raise err.AirbyteConnectionInactiveException(
            f"Please enable the Connection {connection_id} in Airbyte instance."
        )
    elif connection_status == CONNECTION_STATUS_DEPRECATED:
        logger.error(f"Connection {connection_id} is deprecated.")
        raise err.AirbyeConnectionDeprecatedException(
            f"Connection {connection_id} is deprecated."
        )



"""Tasks for updating and fetching Airbyte configurations"""
from prefect import task
from prefect.logging.loggers import get_logger

from prefect_airbyte.client import AirbyteClient


@task
async def export_configuration(
    airbyte_server_host: str = "localhost",
    airbyte_server_port: int = "8000",
    airbyte_api_version: str = "v1",
    timeout: int = 5,
) -> bytearray:

    """
    Task that exports an Airbyte config via `{AIRBYTE_HOST}/api/v1/deployment/export`
    Args:
        airbyte_server_host (str, optional): Hostname of Airbyte server where
            connection is configured. Will overwrite the value provided at init
            if provided.
        airbyte_server_port (str, optional): Port that the Airbyte server is
            listening on, will overwrite the value provided at init if provided.
        airbyte_api_version (str, optional): Version of Airbyte API to use to
            trigger connection sync, will overwrite the value provided at
            init if provided.
        timeout (int): timeout in seconds on the httpx AirbyteClient

    Returns:
        bytearray: `bytearray` containing Airbyte configuration

    Examples:

        Flow that writes the Airbyte configuration as a gzip to a filepath:

        ```python
        import gzip

        from prefect import flow, task
        from prefect_airbyte.configuration import export_configuration

        @task
        def zip_and_write_somewhere(
            airbyte_configuration: bytearray
            somewhere: str = 'my_destination.gz','
        ):
            with gzip.open('my_destination.gz', 'wb') as f:
                    f.write(airbyte_configuration)

        @flow
        def example_export_configuration_flow():

            # Run other tasks and subflows here

            airbyte_config = export_configuration(
                    airbyte_server_host="localhost",
                    airbyte_server_port="8000",
                    airbyte_api_version="v1",
            )

            zip_and_write_somewhere(airbyte_config=airbyte_config)

        example_trigger_sync_flow()
        ```
    """

    logger = get_logger()

    airbyte_base_url = (
        f"http://{airbyte_server_host}:"
        f"{airbyte_server_port}/api/{airbyte_api_version}"
    )

    airbyte = AirbyteClient(logger, airbyte_base_url, timeout=timeout)

    logger.info("Initiating export of Airbyte configuration")
    airbyte_config = await airbyte.export_configuration()

    return airbyte_config
